"""Implementations of algorithms for continuous control."""
import functools
from typing import Optional, Sequence, Tuple

import jax
import jax.numpy as jnp
import numpy as np
import optax

import policy
import value_net
from actor import update as awr_update_actor
from common import Batch, InfoDict, Model, PRNGKey
from critic import update_q, update_v


def target_update(critic: Model, target_critic: Model, tau: float) -> Model:
    new_target_params = jax.tree_multimap(
        lambda p, tp: p * tau + tp * (1 - tau), critic.params, target_critic.params
    )

    return target_critic.replace(params=new_target_params)


@functools.partial(
    jax.jit, static_argnames=("critic_ensemble_size", "critic_training_iterations")
)
def _update_jit(
    rng: PRNGKey,
    actor: Model,
    critic: Model,
    value: Model,
    target_critic: Model,
    batch: Batch,
    discount: float,
    tau: float,
    expectile: float,
    temperature: float,
    critic_ensemble_size: int,
    critic_training_iterations: int,
) -> Tuple[PRNGKey, Model, Model, Model, Model, Model, InfoDict]:

    for it in range(critic_training_iterations):
        key, rng = jax.random.split(rng)
        new_value, value_info = update_v(
            key, target_critic, value, batch, expectile, critic_ensemble_size
        )

        new_critic, critic_info = update_q(critic, new_value, batch, discount)

        new_target_critic = target_update(new_critic, target_critic, tau)
        (critic, target_critic, value) = (new_critic, new_target_critic, new_value)

    key, rng = jax.random.split(rng)
    new_actor, actor_info = awr_update_actor(
        key, actor, target_critic, new_value, batch, temperature
    )

    return (
        rng,
        new_actor,
        new_critic,
        new_value,
        new_target_critic,
        {**critic_info, **value_info, **actor_info},
    )


class Learner(object):
    def __init__(
        self,
        seed: int,
        observations: jnp.ndarray,
        actions: jnp.ndarray,
        actor_lr: float = 3e-4,
        value_lr: float = 3e-4,
        critic_lr: float = 3e-4,
        hidden_dims: Sequence[int] = (256, 256),
        discount: float = 0.99,
        tau: float = 0.005,
        expectile: float = 0.8,
        temperature: float = 0.1,
        dropout_rate: Optional[float] = None,
        max_steps: Optional[int] = None,
        opt_decay_schedule: str = "cosine",
        critic_ensemble_size: int = 2,
        critic_training_iterations: int = 1,
    ):
        """
        An implementation of the version of Soft-Actor-Critic described in https://arxiv.org/abs/1801.01290
        """

        self.expectile = expectile
        self.tau = tau
        self.discount = discount
        self.temperature = temperature
        self.critic_ensemble_size = critic_ensemble_size
        self.critic_training_iterations = critic_training_iterations

        rng = jax.random.PRNGKey(seed)
        rng, actor_key, critic_key, value_key = jax.random.split(rng, 4)

        action_dim = actions.shape[-1]
        actor_def = policy.NormalTanhPolicy(
            hidden_dims,
            action_dim,
            log_std_scale=1e-3,
            log_std_min=-5.0,
            dropout_rate=dropout_rate,
            state_dependent_std=False,
            tanh_squash_distribution=False,
        )

        if opt_decay_schedule == "cosine":
            schedule_fn = optax.cosine_decay_schedule(-actor_lr, max_steps)
            optimiser = optax.chain(
                optax.scale_by_adam(), optax.scale_by_schedule(schedule_fn)
            )
        else:
            optimiser = optax.adam(learning_rate=actor_lr)

        actor = Model.create(actor_def, inputs=[actor_key, observations], tx=optimiser)

        # critic_def = value_net.DoubleCritic(hidden_dims)
        critic_def = value_net.CriticEnsemble(
            hidden_dims, ensemble_size=critic_ensemble_size
        )
        critic = Model.create(
            critic_def,
            inputs=[critic_key, observations, actions],
            tx=optax.adam(learning_rate=critic_lr),
        )

        value_def = value_net.ValueCritic(hidden_dims)
        value = Model.create(
            value_def,
            inputs=[value_key, observations],
            tx=optax.adam(learning_rate=value_lr),
        )

        target_critic = Model.create(
            critic_def, inputs=[critic_key, observations, actions]
        )

        self.actor = actor
        self.critic = critic
        self.value = value
        self.target_critic = target_critic
        self.rng = rng

    def sample_actions(
        self,
        observations: np.ndarray,
        temperature: float = 1.0,
        num_actions_to_sample: int = 1,
        fixed_action_noise: float = -1,
        optimism_parameter: float = 0,
    ) -> Tuple[jnp.ndarray, InfoDict]:
        rng, actions, info = policy.sample_actions(
            self.rng,
            self.actor.apply_fn,
            self.actor.params,
            observations,
            temperature,
            num_actions_to_sample,
            fixed_action_noise,
            self.target_critic.apply_fn,
            self.target_critic.params,
            optimism_parameter,
        )
        self.rng = rng

        actions = np.asarray(actions)
        return (np.clip(actions, -1, 1), info)

    def update(self, batch: Batch, step: int) -> InfoDict:
        critic_training_iterations = 1 if step < 1 else self.critic_training_iterations

        (
            new_rng,
            new_actor,
            new_critic,
            new_value,
            new_target_critic,
            info,
        ) = _update_jit(
            self.rng,
            self.actor,
            self.critic,
            self.value,
            self.target_critic,
            batch,
            self.discount,
            self.tau,
            self.expectile,
            self.temperature,
            self.critic_ensemble_size,
            critic_training_iterations,
        )

        self.rng = new_rng
        self.actor = new_actor
        self.critic = new_critic
        self.value = new_value
        self.target_critic = new_target_critic

        return info
